#!/usr/bin/env python3
# coding=utf-8
# 
# Clinklog: static site generator for linklogs and microblogs
# 2024-01-08 Felix Pleșoianu <https://felix.plesoianu.ro/>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
static site generator for linklogs and microblogs

This script generates static pages from a theme and data file.
"""

import configparser
import html
import urllib.parse
import string
import re
import sys
import os
import os.path
import datetime as dt

try:
	import pytz
except ImportError as e:
	print("Warning: timezone support not found.", file=sys.stderr)

version_string = "Clinklog Lite r1.0"

project = configparser.ConfigParser()
project["CONFIG"] = {
	"title": "My clinklog",
	"desc": "A collection of thoughts and bookmarks",
	"link": "",
	"lang": "",
	"header": "",
	"footer": "",
	"logo": "",
	"theme": "theme",
	"css": "clinklog.css"
}

html_page = ""
html_item = ""
page_css = ""

html_logo = """
 <a href="{site_link:escape}"><img src="{site_logo:escape}"
    alt="Home" class="logo"></a>
"""

rss_page = """<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0">
 <channel>
  <title>{site_title:escape}</title>
  <link>{site_link:escape}</link>
  <description>{site_desc:escape}</description>
  {logo}
  <language>{lang:escape}</language>
  <copyright>{copy:escape}</copyright>
  <generator>{generator:escape}</generator>
  <lastBuildDate>{build_date:escape}</lastBuildDate>
  <docs>https://cyber.harvard.edu/rss/rss.html</docs>

  {rss_items}
 </channel>
</rss>
"""

rss_image = """
  <image>
   <url>{site_logo:escape}</url>
   <title>{site_title:escape}</title>
   <link>{site_link:escape}</link>
  </image>
"""

rss_item = """
   <item>
    <title>{title:escape}</title>
    <link>{link:escape}</link>
    <description>{desc:escape}</description>
    <pubDate>{pubdate:escape}</pubDate>
    <guid isPermaLink="true">{permalink:escape}</guid>
    {categories}
   </item>
"""

rfc822_date = "%a, %d %b %Y %H:%M:%S %Z"
blog_timezone = None

# Inspired by https://github.com/ebrehault/superformatter
class WebFormatter(string.Formatter):
	"Adds an HTML `escape` formatting code to Python."
	def format_field(self, value, spec):
		if spec == "escape":
			return html.escape(value)
		elif spec == "url":
			return urllib.parse.quote(value)
		else:
			return super().format_field(value, spec)

def theme_file(name):
	return os.path.join(project["CONFIG"]["theme"], name)

def load_file(name):
	try:
		with open(name) as f:
			return f.read()
	except OSError as e:
		print(e, file=sys.stderr)
		return None

def display_date(source):
	if source.hour > 0 or source.minute > 0:
		return source.isoformat(" ", "minutes")
	else:
		return source.date().isoformat()

def parse_item(slug, item):
	"Return item with date/tags as structured data, and slug added."
	parsed = dict(item)
	parsed["slug"] = slug
	parsed["pubdate"] = dt.datetime.fromisoformat(item["pubdate"])
	parsed["pubdate"] = parsed["pubdate"].astimezone(blog_timezone)
	parsed["datestr"] = display_date(parsed["pubdate"])
	if "tags" in parsed:
		parsed["tags"] = [i.strip()
			for i in parsed["tags"].split(",")]
	else:
		parsed["tags"] = []
	return parsed

def parse_note(slug, note):
	"Return note with a default link if not given, and slug added."
	parsed = dict(note)
	parsed["slug"] = slug
	if not parsed.get("link"):
		parsed["link"] = "notes.html#note-" + slug
	return parsed

def permalink(page, settings):
	if settings.get("link"):
		return settings["link"].rstrip("/") + "/" + page
	else:
		return page

def format_tags(seq):
	return " ".join('#' + format_slug(i) for i in seq)

def format_slug(text):
	return "-".join(re.split("\W+", text)).lower()

def format_category(text):
	return "<category>{0}</category>".format(html.escape(text))

def html_fields(item):
	"Map item data to HTML template fields."
	anchor = "item-{0}".format(item["slug"])
	page_link = "#{}".format(anchor)
	taglist = format_tags(item.get("tags"))
	
	return {
		"anchor": anchor,
		"link": item.get("link", "#" + anchor),
		"title": item.get("title") or item.get("link") or "",
		"desc": item.get("desc", ""),
		"pubdate": item["datestr"],
		"permalink": page_link,
		"tags": "<p>{0}</p>".format(taglist)
	}

def note_fields(note):
	"Map note data to HTML template fields."
	anchor = "note-{0}".format(note["slug"])
	page_link = "notes.html" + anchor
	
	return {
		"anchor": anchor,
		"link": note.get("link", "#" + anchor),
		"title": note.get("title") or note.get("link") or "",
		"desc": note.get("desc"),
		"pubdate": "",
		"permalink": page_link,
		"tags": ""
	}

def rss_fields(item):
	"Map item data to RSS template fields."
	anchor = "#item-{0}".format(item["slug"])
	perma = permalink(anchor, project["CONFIG"])
	if item.get("tags"):
		categories = [format_category(i) for i in item["tags"]]
	else:
		categories = []
	utcdate = item["pubdate"].astimezone(dt.timezone.utc)
	
	return {
		"link": item.get("link", ""),
		"title": item.get("title", ""),
		"desc": item.get("desc", ""),
		"pubdate": utcdate.strftime(rfc822_date),
		"permalink": perma,
		"categories": "\n    ".join(categories)
	}

def menu_link(text, url, target=""):
	if target == "":
		return '<a href="{0}">{1}</a>'.format(
			html.escape(url), html.escape(text, False))
	else:
		return '<a href="{0}" target="{1}">{2}</a>'.format(
			html.escape(url),
			html.escape(target),
			html.escape(text, False))

def make_menu(items, t=""):
	if len(items) == 0: return ""
	items = ("  <li>{0}</li>".format(menu_link(k, v, t))
		for k, v in items)
	return "<ul>\n{0}\n </ul>".format("\n".join(items))

def format_logo(link, logo):
	wf = WebFormatter()
	return wf.format(html_logo,
		site_link=html.escape(link, True),
		site_logo=html.escape(logo, True))

def format_page(title, items, menu, pclass=""):
	wf = WebFormatter()
	settings = project["CONFIG"]
	link = settings.get("link", "index.html")
	target = settings.get("target", "")
	if settings.get("logo"):
		logo_string = format_logo(link, settings["logo"])
	else:
		logo_string = ""
	return wf.format(html_page,
		site_title=settings.get("title", ""),
		site_link=link,
		site_desc=settings.get("desc", ""),
		site_css=settings.get("css", ""),
		site_menu=make_menu(menu, target),
		logo=logo_string,
		lang=settings.get("lang", ""),
		header=settings.get("header", ""),
		footer=settings.get("footer", ""),
		generator=version_string,
		page_title=title,
		page_items=items,
		page_class=pclass)

def make_index(items, menu):
	wf = WebFormatter()
	last_heading = ""
	text = ""
	for i in items:
		month = i["pubdate"].strftime("%B %Y")
		if month != last_heading:
			text += "<h2>" + month + "</h2>\n"
			last_heading = month
		text += wf.format(html_item, **html_fields(i))
	return format_page("Recent", text, menu, "index")

def make_notes(notes, menu):
	wf = WebFormatter()
	notes = (wf.format(html_item, **note_fields(i)) 
		for i in notes if i.get("desc"))
	text = "<h2>Notes</h2>" + "\n".join(notes)
	return format_page("Notes", text, menu, "page")

def make_rss(items):
	wf = WebFormatter()
	settings = project["CONFIG"]
	items = (wf.format(rss_item, **rss_fields(i)) for i in items)
	link = settings.get("link", "")
	now = dt.datetime.now(dt.timezone.utc)
	if settings.get("logo"):
		logo_string = wf.format(rss_image,
			site_title=settings.get("title", ""),
			site_link=link,
			site_logo=settings["logo"])
	else:
		logo_string = ""
	return wf.format(rss_page,
		site_title=settings.get("title", ""),
		site_link=link,
		site_desc=settings.get("desc", ""),
		logo=logo_string,
		lang=settings.get("lang", ""),
		copy=settings.get("copy", ""),
		generator=version_string,
		build_date=now.strftime(rfc822_date),
		rss_items="\n".join(items))

if __name__ == "__main__":
	db_filename = os.getenv("CLINKLOGDB", "clinklog.ini")	
	exp_folder = os.getenv("CLINKLOGEXP", os.getcwd())
	try:
		exp_limit = int(os.getenv("CLINKLOGNUM", 100))
	except ValueError as e:
		exp_limit = 100
	
	import argparse
	import shlex

	cmdline = argparse.ArgumentParser(
		description="Generate a static linklog or microblog.",
		epilog="A theme directory is needed for this to work.")
	cmdline.add_argument("-v", "--version", action="version",
		version=version_string)

	cmdline.add_argument("-q", "--quiet", action="store_true",
		help="don't print confirmation messages")
	cmdline.add_argument("-t", "--theme", metavar="DIR",
		help="override theme folder set in project config")
	cmdline.add_argument("-o", "--output",
		metavar="DIR", default=exp_folder,
		help="export to given folder (default: current)")
	cmdline.add_argument("-n", "--limit",
		type=int, default=exp_limit, metavar="N",
		help="max. items on front page (default: 100)")

	cmdline.add_argument("files", nargs='*',
		default=shlex.split(db_filename),
		help="blog file(s) to work on (default: clinklog.ini)")
	
	args = cmdline.parse_args()
	
	# Load data files manually to catch and report errors.
	try:
		files = args.files
		for i in files:
			with open(i) as f:
				project.read_file(f, source=i)
			if not args.quiet:
				print("Loaded", i)
	except OSError as e:
		cmdline.exit(1, str(e) + "\n")
	except configparser.Error as e:
		cmdline.exit(1, str(e) + "\n")
	
	if "timezone" in project["CONFIG"] and "pytz" in sys.modules:
		try:
			blog_timezone = pytz.timezone(
				project["CONFIG"]["timezone"])
		except pytz.exceptions.UnknownTimeZoneError as e:
			print("Warning: bad timezone",
				project["CONFIG"]["timezone"],
				file=sys.stderr)
	
	if args.theme:
		project["CONFIG"]["theme"] = args.theme
	
	html_page = load_file(theme_file("page.html"))
	html_item = load_file(theme_file("item.html"))
	page_css = load_file(theme_file(project["CONFIG"]["css"]))
	
	if not all([html_page, html_item, page_css]):
		cmdline.exit(2, "Missing theme files, stopping.\n")
	elif not args.quiet:
		print("Theme loaded from:", project["CONFIG"]["theme"])

	all_items = []
	all_notes = []
	for i in project:
		if "pubdate" in project[i]:
			try:
				all_items.append(parse_item(i, project[i]))
			except ValueError as e:
				print("Warning: bad date in item", i,
					end='', file=sys.stderr)
				print("(", str(e), ")", file=sys.stderr)
		elif i not in ("CONFIG", "DEFAULT"):
			all_notes.append(parse_note(i, project[i]))
	all_items.sort(key=lambda i: i["pubdate"], reverse=True)
	# all_notes.sort(key=lambda i: i["slug"], reverse=False)
	
	recent = all_items[0:args.limit]
	
	menu = [(i["title"], i["link"]) for i in all_notes]

	try:
		os.makedirs(args.output, exist_ok=True)
		index_file = os.path.join(args.output, "index.html")
		with open(index_file, "w") as f:
			f.write(make_index(recent, menu))
		notes_file = os.path.join(args.output, "notes.html")
		with open(notes_file, "w") as f:
			f.write(make_notes(all_notes, menu))
		rss_file = os.path.join(args.output, "recent.xml")
		with open(rss_file, "w") as f:
			f.write(make_rss(recent))
		style_file = os.path.join(
			args.output, project["CONFIG"]["css"])
		with open(style_file, "w") as f:
			f.write(page_css)
	except OSError as e:
		cmdline.exit(3, str(e) + "\n")

	print("Blog exported to:", args.output)
